import math

from torch import nn

import ganetic.basemodel as basemodel


class Generator(basemodel.Generator):
    r"""Deep Convolutional Generative Adversarial Network (DCGAN) Generator.
    Parameters
    ----------
    nz : int, default 100
        Size of the latent z vector.
    nc : int, default 3
        Number of channels in the output image.
    ngf: int, default 64
        Size of feature maps in generator.
    out_size: int, default 64
        Size of the output image.
    activation: torch.nn.Module, default nn.ReLU(True)
        Activation function to use in the generator.
    last_activation: torch.nn.Module, default nn.Tanh()
        Activation function to use in the last layer of the generator.
    """

    def __init__(
        self,
        nz=100,
        nc=3,
        ngf=64,
        out_size=64,
        activation=nn.ReLU(True),
        last_activation=nn.Tanh(),
    ):
        super(Generator, self).__init__()
        if out_size < 16 and math.ceil(math.log2(out_size)) != math.floor(
            math.log2(out_size)
        ):
            raise Exception("out_size must be a power of 2 and greater than 16")
        total_repeats = out_size.bit_length() - 4
        model = []
        _ngf = ngf * 2**total_repeats
        model += [
            nn.Sequential(
                nn.ConvTranspose2d(nz, _ngf, 4, 1, 0, bias=False),
                nn.BatchNorm2d(_ngf),
                activation,
            )
        ]
        for _ in range(total_repeats):
            _ngf //= 2
            model += [
                nn.Sequential(
                    nn.ConvTranspose2d(_ngf * 2, _ngf, 4, 2, 1, bias=False),
                    nn.BatchNorm2d(_ngf),
                    activation,
                )
            ]
        model += [
            nn.Sequential(
                nn.ConvTranspose2d(ngf, nc, 4, 2, 1, bias=False), last_activation
            )
        ]
        self.model = nn.Sequential(*model)
        self._weights_init(self.model)

    def forward(self, x):
        return self.model(x.view(x.size(0), x.size(1), 1, 1))


class Discriminator(basemodel.Discriminator):
    r"""Deep Convolutional Generative Adversarial Network (DCGAN) Discriminator.
    Parameters
    ----------
    nc : int, default 3
        Number of channels in the input image.
    ndf: int, default 64
        Size of feature maps in discriminator.
    in_size: int, default 64
        Size of the input image.
    activation: torch.nn.Module, default nn.LeakyReLU(0.2, inplace=True)
        Activation function to use in the discriminator.
    last_activation: torch.nn.Module, default nn.Sigmoid()
        Activation function to use in the last layer of the discriminator.
    """

    def __init__(
        self,
        nc=3,
        ndf=64,
        in_size=64,
        activation=nn.LeakyReLU(0.2, inplace=True),
        last_activation=nn.Sigmoid(),
    ):
        super(Discriminator, self).__init__()
        if in_size < 16 and math.ceil(math.log2(in_size)) != math.floor(
            math.log2(in_size)
        ):
            raise Exception("in_size must be a power of 2 and greater than 16")
        total_repeats = in_size.bit_length() - 4
        model = []
        _ndf = ndf
        model += [nn.Sequential(nn.Conv2d(nc, _ndf, 4, 2, 1, bias=False), activation)]
        for _ in range(total_repeats):
            model += [
                nn.Sequential(
                    nn.Conv2d(_ndf, _ndf * 2, 4, 2, 1, bias=False),
                    nn.BatchNorm2d(_ndf * 2),
                    activation,
                )
            ]
            _ndf *= 2
        model += [
            nn.Sequential(nn.Conv2d(_ndf, 1, 4, 1, 0, bias=False), last_activation)
        ]
        self.model = nn.Sequential(*model)
        self._weights_init(self.model)

    def forward(self, x):
        return self.model(x).view(-1, 1).squeeze(1)



import math

import torch
from torch import nn

import ganetic.basemodel as basemodel

from ganetic.layers import ResidualBlock, UpsampleBlock

# TODO[#1]: Make Generator and Discriminator
# classes inherit from a base class


class Generator(basemodel.Generator):
    r"""
    Parameters
    ----------
    scale_factor : int
        The scale factor for the super resolution.
    nci : int
        The number of channels in the input tensor.
    nco : int
        The number of channels in the output tensor.
    ngf : int, default=64
        The number of filters in the generator.
    n_res : int, default=5
        The number of residual ResidualBlock in the generator.
    last_activation_fn : lambda, default=(nn.Tanh()+1)/2
    """

    def __init__(
            self, scale_factor, nci=3, nco=3, ngf=64, n_res=5, last_activation_fn=lambda x: (nn.tanh(x) + 1) / 2):
        super(Generator, self).__init__()
        upsample_block_num = int(math.log(scale_factor, 2))

        self.block1 = nn.Sequential(
            nn.Conv2d(nci, ngf, kernel_size=9, padding=4), nn.PReLU()
        )
        block2 = [ResidualBlock(ngf) for _ in range(n_res)]
        self.block2 = nn.Sequential(*block2)
        self.block3 = nn.Sequential(
            nn.Conv2d(ngf, ngf, kernel_size=3, padding=1), nn.BatchNorm2d(ngf)
        )
        block4 = [UpsampleBlock(ngf, 2) for _ in range(upsample_block_num)]
        block4.append(nn.Conv2d(ngf, nco, kernel_size=9, padding=4))
        self.block4 = nn.Sequential(*block4)
        self.last_activation_fn = last_activation_fn

    def forward(self, x):
        x = self.block1(x)
        block2 = x
        block2 = self.block2(block2)
        block3 = self.block3(block2) + x
        block4 = self.block4(block3)
        return self.last_activation_fn(block4)


class Discriminator(basemodel.Discriminator):
    r"""
    Parameters
    ----------
    input_shape : tuple
        The shape of the input tensor.
    ndf : int, default=64
        The number of filters in the discriminator.
    negative_slope : float, default=0.2
        The negative slope of the leaky relu.
    """

    def __init__(
        self,
        input_shape,
        ndf=64,
        negative_slope=0.2,
    ):
        super(Discriminator, self).__init__()
        self.block1 = nn.Sequential(
            nn.Conv2d(input_shape[0], ndf, kernel_size=3, stride=1, padding=1),
            nn.LeakyReLU(negative_slope=negative_slope, inplace=True),
        )
        self.block2 = nn.Sequential(
            nn.Conv2d(ndf, ndf, kernel_size=3, stride=2, padding=1),
            nn.BatchNorm2d(ndf),
            nn.LeakyReLU(negative_slope=negative_slope, inplace=True),
        )
        self.block3 = nn.Sequential(
            nn.Conv2d(ndf, ndf * 2, kernel_size=3, stride=1, padding=1),
            nn.BatchNorm2d(ndf * 2),
            nn.LeakyReLU(negative_slope=negative_slope, inplace=True),
        )
        self.block4 = nn.Sequential(
            nn.Conv2d(ndf * 2, ndf * 2, kernel_size=3, stride=2, padding=1),
            nn.BatchNorm2d(ndf * 2),
            nn.LeakyReLU(negative_slope=negative_slope, inplace=True),
        )
        self.block5 = nn.Sequential(
            nn.Conv2d(ndf * 2, ndf * 4, kernel_size=3, stride=1, padding=1),
            nn.BatchNorm2d(ndf * 4),
            nn.LeakyReLU(negative_slope=negative_slope, inplace=True),
        )
        self.block6 = nn.Sequential(
            nn.Conv2d(ndf * 4, ndf * 4, kernel_size=3, stride=2, padding=1),
            nn.BatchNorm2d(ndf * 4),
            nn.LeakyReLU(negative_slope=negative_slope, inplace=True),
        )
        self.block7 = nn.Sequential(
            nn.Conv2d(ndf * 4, ndf * 8, kernel_size=3, stride=1, padding=1),
            nn.BatchNorm2d(ndf * 8),
            nn.LeakyReLU(negative_slope=negative_slope, inplace=True),
        )
        self.block8 = nn.Sequential(
            nn.Conv2d(ndf * 8, ndf * 8, kernel_size=3, stride=2, padding=1),
            nn.BatchNorm2d(ndf * 8),
            nn.LeakyReLU(negative_slope=negative_slope, inplace=True),
        )
        with torch.no_grad():
            linear_size = (
                self.block8(
                    self.block7(
                        self.block6(
                            self.block5(
                                self.block4(
                                    self.block3(
                                        self.block2(
                                            self.block1(torch.zeros(
                                                1, *(input_shape)))
                                        )
                                    )
                                )
                            )
                        )
                    )
                )
                .view(1, -1)
                .size(1)
            )
        self.block9 = nn.Sequential(
            nn.Linear(linear_size, 1024),
            nn.LeakyReLU(negative_slope=negative_slope, inplace=True),
        )
        self.block10 = nn.Sequential(nn.Linear(1024, 1), nn.Sigmoid())

    def forward(self, x):
        x = self.block1(x)
        x = self.block2(x)
        x = self.block3(x)
        x = self.block4(x)
        x = self.block5(x)
        x = self.block6(x)
        x = self.block7(x)
        x = self.block8(x)
        x = x.view(x.size(0), -1)
        x = self.block9(x)
        x = self.block10(x)
        return x



from torch import nn
from .dcgan import Generator as DCGANGenerator, Discriminator as DCGANDiscriminator
import torch


class Generator(DCGANGenerator):
    r"""Conditional Generative Adversarial Network (CGAN) Generator.
    Parameters
    ----------
    n_classes : int
        Number of classes.
    nz : int, default 100
        Size of the latent z vector.
    nc : int, default 3
        Number of channels in the output image.
    ngf: int, default 64
        Size of feature maps in generator.
    out_size: int, default 64
        Size of the output image.
    activation: torch.nn.Module, default nn.ReLU(True)
        Activation function to use in the generator.
    last_activation: torch.nn.Module, default nn.Tanh()
        Activation function to use in the last layer of the generator.
    """

    def __init__(
        self,
        n_classes,
        nz=100,
        nc=3,
        ngf=64,
        out_size=64,
        activation=nn.ReLU(True),
        last_activation=nn.Tanh(),
    ):
        super(Generator, self).__init__(
            nz + n_classes, nc, ngf, out_size, activation, last_activation
        )
        self.n_classes = n_classes
        self.nz = nz
        self.label_embedding = nn.Embedding(n_classes, n_classes)

    def forward(self, x, labels):
        label_output = self.label_embedding(labels)
        return super(Generator, self).forward(torch.cat((x, label_output), 1))


class Discriminator(DCGANDiscriminator):
    r"""
    Parameters
    ----------
    n_classes : int
        Number of classes.
    nc : int, default 3
        Number of channels in the input image.
    ndf: int, default 64
        Size of feature maps in discriminator.
    in_size: int, default 64
        Size of the input image.
    activation: torch.nn.Module, default nn.LeakyReLU(0.2, inplace=True)
        Activation function to use in the discriminator.
    last_activation: torch.nn.Module, default nn.Sigmoid()
        Activation function to use in the last layer of the discriminator.
    """

    def __init__(
        self,
        n_classes,
        nc=3,
        ndf=64,
        in_size=64,
        activation=nn.LeakyReLU(0.2, inplace=True),
        last_activation=nn.Sigmoid(),
    ):
        super(Discriminator, self).__init__(
            nc + n_classes, ndf, in_size, activation, last_activation
        )
        self.n_classes = n_classes
        self.label_embedding = nn.Embedding(n_classes, n_classes)

    def forward(self, x, labels):
        label_output = self.label_embedding(labels)
        label = (
            label_output.unsqueeze(2)
            .unsqueeze(3)
            .expand(label_output.size(0), label_output.size(1), x.size(2), x.size(3))
        )
        return super(Discriminator, self).forward(torch.cat((x, label), 1))



from torch import nn


class Generator(nn.Module):
    r"""
    Base class for all Generators.
    """

    def __init__(self):
        super(Generator, self).__init__()

    def _weights_init(self, m):
        r"""
        Default weights initialization for all Generators.
        """
        classname = m.__class__.__name__
        if classname.find("Conv") != -1:
            nn.init.normal_(m.weight.data, 0.0, 0.02)
        elif classname.find("BatchNorm") != -1:
            nn.init.normal_(m.weight.data, 1.0, 0.02)
            nn.init.constant_(m.bias.data, 0)


class Discriminator(nn.Module):
    r"""
    Base class for all Discriminators.
    """

    def __init__(self):
        super(Discriminator, self).__init__()

    def _weights_init(self, m):
        r"""
        Default weights initialization for all Discriminators.
        """
        classname = m.__class__.__name__
        if classname.find("Conv") != -1:
            nn.init.normal_(m.weight.data, 0.0, 0.02)
        elif classname.find("BatchNorm") != -1:
            nn.init.normal_(m.weight.data, 1.0, 0.02)
            nn.init.constant_(m.bias.data, 0)



import torch
from torch import nn

# TODO[#1]: Make Generator and Discriminator
# classes inherit from a base class
# TODO[#2]: Remove normal_init function and use
# the one in basemodel.py


def normal_init(m, mean, std):
    if isinstance(m, nn.ConvTranspose2d) or isinstance(m, nn.Conv2d):
        m.weight.data.normal_(mean, std)
        m.bias.data.zero_()


class ConvBlock(torch.nn.Module):
    r"""
    Convolutional-BatchNorm-ReLU layer with k filters.
    Parameters
    ----------
    in_channels : int
        The number of channels in the input tensor.
    out_channels : int
        The number of channels in the output tensor.
    kernel_size : int, default=4
        The size of the convolutional kernel.
    stride : int, default=2
        The stride of the convolution.
    padding : int, default=1
        The padding of the convolution.
    activation : bool, default=True
        Whether to use a ReLU activation.
    batch_norm : bool, default=True
        Whether to use a batch normalization layer.
    """

    def __init__(
        self,
        in_channels,
        out_channels,
        kernel_size=4,
        stride=2,
        padding=1,
        activation=True,
        batch_norm=True,
    ):
        super(ConvBlock, self).__init__()
        self.conv = nn.Conv2d(
            in_channels,
            out_channels,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
        )
        self.activation = activation
        self.batch_norm = batch_norm
        if self.activation:
            self.relu = nn.LeakyReLU(0.2, inplace=True)
        if self.batch_norm:
            self.bn = nn.BatchNorm2d(out_channels)

    def forward(self, x):
        if self.activation:
            x = self.relu(x)
        x = self.conv(x)
        if self.batch_norm:
            x = self.bn(x)
        return x


class DeConvBlock(torch.nn.Module):
    r"""
    Convolutional-BatchNorm-Dropout-ReLU layer with 50% dropout.
    Parameters
    ----------
    in_channels : int
        The number of channels in the input tensor.
    out_channels : int
        The number of channels in the output tensor.
    kernel_size : int, default=4
        The size of the convolutional kernel.
    stride : int, default=2
        The stride of the convolution.
    padding : int, default=1
        The padding of the convolution.
    batch_norm : bool, default=True
        Whether to use a batch normalization layer.
    dropout : bool, default=False
        Whether to use a dropout layer.
    """

    def __init__(
        self,
        in_channels,
        out_channels,
        kernel_size=4,
        stride=2,
        padding=1,
        batch_norm=True,
        dropout=False,
    ):
        super(DeConvBlock, self).__init__()
        self.deconv = nn.ConvTranspose2d(
            in_channels,
            out_channels,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
        )
        self.batch_norm = batch_norm
        self.dropout = dropout
        if self.batch_norm:
            self.bn = nn.BatchNorm2d(out_channels)
        if self.dropout:
            self.drop = nn.Dropout(0.5)
        self.relu = nn.ReLU(inplace=True)

    def forward(self, x):
        x = self.relu(x)
        x = self.deconv(x)
        if self.batch_norm:
            x = self.bn(x)
        if self.dropout:
            x = self.drop(x)
        return x


class Generator(torch.nn.Module):
    r"""
    Generator network for pix2pix.
    Parameters
    ----------
    nci : int
        The number of channels in the input tensor.
    nco : int, default=None
        The number of channels in the output tensor.
        If None, nco = nci.
    ngf : int, default=64
        The number of filters in the generator.
    """

    def __init__(self, nci, nco=None, ngf=64):
        super(Generator, self).__init__()
        if nco is None:
            nco = nci
        # U-Net encoder
        self.conv1 = ConvBlock(
            in_channels=nci, out_channels=ngf, activation=False, batch_norm=False
        )
        self.conv2 = ConvBlock(in_channels=ngf, out_channels=ngf * 2)
        self.conv3 = ConvBlock(in_channels=ngf * 2, out_channels=ngf * 4)
        self.conv4 = ConvBlock(in_channels=ngf * 4, out_channels=ngf * 8)
        self.conv5 = ConvBlock(in_channels=ngf * 8, out_channels=ngf * 8)
        self.conv6 = ConvBlock(in_channels=ngf * 8, out_channels=ngf * 8)
        self.conv7 = ConvBlock(in_channels=ngf * 8, out_channels=ngf * 8)
        self.conv8 = ConvBlock(
            in_channels=ngf * 8, out_channels=ngf * 8, batch_norm=False
        )

        # U-Net decoder
        self.deconv1 = DeConvBlock(
            in_channels=ngf * 8, out_channels=ngf * 8, dropout=True
        )
        self.deconv2 = DeConvBlock(
            in_channels=ngf * 8 * 2, out_channels=ngf * 8, dropout=True
        )
        self.deconv3 = DeConvBlock(
            in_channels=ngf * 8 * 2, out_channels=ngf * 8, dropout=True
        )
        self.deconv4 = DeConvBlock(in_channels=ngf * 8 * 2, out_channels=ngf * 8)
        self.deconv5 = DeConvBlock(in_channels=ngf * 8 * 2, out_channels=ngf * 4)
        self.deconv6 = DeConvBlock(in_channels=ngf * 4 * 2, out_channels=ngf * 2)
        self.deconv7 = DeConvBlock(in_channels=ngf * 2 * 2, out_channels=ngf)
        self.deconv8 = DeConvBlock(
            in_channels=ngf * 2, out_channels=nco, batch_norm=False
        )

    def weight_init(self, mean, std):
        for m in self._modules:
            normal_init(self._modules[m], mean, std)

    def forward(self, x):
        e1 = self.conv1(x)
        e2 = self.conv2(e1)
        e3 = self.conv3(e2)
        e4 = self.conv4(e3)
        e5 = self.conv5(e4)
        e6 = self.conv6(e5)
        e7 = self.conv7(e6)
        e8 = self.conv8(e7)

        d1 = self.deconv1(e8)
        d1 = torch.cat((d1, e7), dim=1)
        d2 = self.deconv2(d1)
        d2 = torch.cat((d2, e6), dim=1)
        d3 = self.deconv3(d2)
        d3 = torch.cat((d3, e5), dim=1)
        d4 = self.deconv4(d3)
        d4 = torch.cat((d4, e4), dim=1)
        d5 = self.deconv5(d4)
        d5 = torch.cat((d5, e3), dim=1)
        d6 = self.deconv6(d5)
        d6 = torch.cat((d6, e2), dim=1)
        d7 = self.deconv7(d6)
        d7 = torch.cat((d7, e1), dim=1)
        d8 = self.deconv8(d7)
        return torch.tanh(d8)


class Discriminator(torch.nn.Module):
    r"""
    Discriminator network for pix2pix.
    Parameters
    ----------
    nci : int
        The number of channels in the input tensor.
    nco : int, default=1
        The number of channels in the output tensor.
    ndf : int, default=64
        The number of filters in the discriminator.
    """

    def __init__(self, nci, ndf=64):
        super(Discriminator, self).__init__()
        self.conv1 = ConvBlock(
            in_channels=nci * 2, out_channels=ndf, activation=False, batch_norm=False
        )
        self.conv2 = ConvBlock(in_channels=ndf, out_channels=ndf * 2)
        self.conv3 = ConvBlock(in_channels=ndf * 2, out_channels=ndf * 4)
        self.conv4 = ConvBlock(
            in_channels=ndf * 4, out_channels=ndf * 8, stride=1, padding=1
        )
        self.conv5 = ConvBlock(
            in_channels=ndf * 8, out_channels=1, stride=1, batch_norm=False
        )

    def weight_init(self, mean, std):
        for m in self._modules:
            normal_init(self._modules[m], mean, std)

    def forward(self, x, y):
        x = torch.cat((x, y), dim=1)
        x = self.conv1(x)
        x = self.conv2(x)
        x = self.conv3(x)
        x = self.conv4(x)
        x = self.conv5(x)
        return torch.sigmoid(x)



import io

from setuptools import find_packages, setup

with io.open("README.md", "r", encoding="utf-8") as f:
    long_description = "\n" + f.read()

setup(
    name="GANetic",
    packages=find_packages(exclude=["docs", "tests*", "examples"]),
    version="0.0.8",
    license="MIT",
    description="A Collection of GANs - PyTorch",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    author="Geo Jolly",
    author_email="geojollyc@gmail.com",
    url="https://github.com/kingjuno/GANetic",
    keywords=[
        "GAN",
        "Generative Adversarial Networks",
        "Deep Learning",
        "PyTorch",
        "GANetic",
        "Generative Artificial Intelligence",
    ],
    install_requires=[
        "torch>=1.10",
    ],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.6",
    ],
)



import torch.nn as nn


class ResidualBlock(nn.Module):
    def __init__(self, channels):
        super(ResidualBlock, self).__init__()
        self.conv1 = nn.Conv2d(channels, channels, kernel_size=3, padding=1)
        self.bn1 = nn.BatchNorm2d(channels)
        self.prelu = nn.PReLU()
        self.conv2 = nn.Conv2d(channels, channels, kernel_size=3, padding=1)
        self.bn2 = nn.BatchNorm2d(channels)

    def forward(self, x):
        _x = x
        x = self.conv1(x)
        x = self.bn1(x)
        x = self.prelu(x)
        x = self.conv2(x)
        x = self.bn2(x)
        return x + _x


class UpsampleBlock(nn.Module):
    def __init__(self, in_channels, up_scale):
        super(UpsampleBlock, self).__init__()
        self.conv = nn.Conv2d(
            in_channels, in_channels * up_scale**2, kernel_size=3, padding=1
        )
        self.pixel_shuffle = nn.PixelShuffle(up_scale)
        self.prelu = nn.PReLU()

    def forward(self, x):
        x = self.conv(x)
        x = self.pixel_shuffle(x)
        x = self.prelu(x)
        return x



from torch import nn
from torch.nn import functional as F

from ganetic import basemodel


class ResidualBlock(nn.Module):
    def __init__(self, channels):
        super(ResidualBlock, self).__init__()
        self.block = nn.Sequential(
            nn.ReflectionPad2d(1),
            nn.Conv2d(channels, channels, kernel_size=3, padding=0),
            nn.InstanceNorm2d(channels),
            nn.ReLU(inplace=True),
            nn.ReflectionPad2d(1),
            nn.Conv2d(channels, channels, kernel_size=3, padding=0),
            nn.InstanceNorm2d(channels),
        )

    def forward(self, x):
        _x = x
        x = self.block(x)
        return x + _x


class Generator(basemodel.Generator):
    """
    CycleGAN Generator
    Paper: https://arxiv.org/abs/1703.10593
    Architecture:
        Let C7s1-k denote a 7×7 Convolution-InstanceNorm-ReLU
        layer with k filters and stride 1, dk denotes a 3×3
        Convolution-InstanceNorm-ReLU layer with k filter and
        stride 2, Rk denotes a residual block that contains two
        3×3 convolutional layers with the same number of filters
        on both layer, uk denotes a 3×3 fractional-strided-Covolution-
        InstanceNorm-ReLU layer with k filters and stride 1/2,

        C7s1-64 => d128 => d256 => R256 => R256
        => R256 => R256 => R256 => R256 => R256
        => R256 => R256 => u128 => u64 => C7s1-3

        (with 9 residual blocks)
    Parameters
    ----------
    nci : int, default 3
        Number of channels in the input image.
    nco : int, default 3
        Number of channels in the output image.
    ngf : int, default 64
        Size of the feature maps in the generator.
    no_of_residual_blocks : int, default 9
        Number of residual blocks in the generator.
    activation : torch.nn.Module, default nn.ReLU(True)
        Activation function to use in the generator.
    last_activation : torch.nn.Module, default nn.Tanh()
        Activation function to use in the last layer of the generator.
    """

    def __init__(
        self,
        nci=3,
        nco=3,
        ngf=64,
        no_of_residual_blocks=9,
        activation=nn.ReLU(True),
        last_activation=nn.Tanh(),
    ):
        super(Generator, self).__init__()
        model = [
            nn.ReflectionPad2d(3),
            nn.Conv2d(nci, ngf, kernel_size=7, padding=0),
            nn.InstanceNorm2d(ngf),
            activation,
        ]

        # Downsampling
        for i in range(2):
            model += [
                nn.Conv2d(
                    ngf * 2**i, ngf * 2 ** (i + 1), kernel_size=3, stride=2, padding=1
                ),
                nn.InstanceNorm2d(ngf * 2 ** (i + 1)),
                activation,
            ]

        # Residual blocks
        for _ in range(no_of_residual_blocks):
            model += [ResidualBlock(ngf * 4)]

        # Upsampling
        for i in range(2):
            model += [
                nn.ConvTranspose2d(
                    ngf * 2 ** (2 - i),
                    ngf * 2 ** (1 - i),
                    kernel_size=3,
                    stride=2,
                    padding=1,
                    output_padding=1,
                ),
                nn.InstanceNorm2d(ngf * 2 ** (1 - i)),
                activation,
            ]

        model += [
            nn.ReflectionPad2d(3),
            nn.Conv2d(ngf, nco, kernel_size=7, padding=0),
            last_activation,
        ]

        self.model = nn.Sequential(*model)
        self._weights_init(self.model)

    def forward(self, x):
        return self.model(x)


class Discriminator(basemodel.Discriminator):
    """
    CycleGAN Discriminator
    Paper: https://arxiv.org/abs/1703.10593
    Architecture:
        C64-C128-C256-C512

        (with 70x70 PatchGAN)
    Parameters
    ----------
    nci : int, default 3
        Number of channels in the input image.
    ndf : int, default 64
        Size of the feature maps in the discriminator.
    activation : torch.nn.Module, default nn.LeakyReLU(0.2, True)
        Activation function to use in the discriminator.
    last_activation : torch.nn.Module, default nn.Sigmoid()
        Activation function to use in the last layer of the discriminator.
    """

    def __init__(
        self,
        nci=3,
        ndf=64,
        activation=nn.LeakyReLU(0.2, True),
        last_activation=nn.Sigmoid(),
    ):
        super(Discriminator, self).__init__()
        model = [
            nn.Conv2d(nci, ndf, kernel_size=4, stride=2, padding=1),
            activation,
        ]

        for i in range(1, 3):
            model += [
                nn.Conv2d(
                    ndf * 2 ** (i - 1), ndf * 2**i, kernel_size=4, stride=2, padding=1
                ),
                nn.InstanceNorm2d(ndf * 2**i),
                activation,
            ]

        model += [
            nn.Conv2d(ndf * 2**2, ndf * 2**3, kernel_size=4, stride=1, padding=1),
            nn.InstanceNorm2d(ndf * 2**3),
            activation,
            nn.Conv2d(ndf * 2**3, 1, kernel_size=4, stride=1, padding=1),
            last_activation,
        ]

        self.model = nn.Sequential(*model)
        self._weights_init(self.model)

    def forward(self, x):
        x = self.model(x)
        return F.avg_pool2d(x, x.size()[2:]).view(-1)
